# Copyright 2020 InterDigital Communications, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# https://github.com/InterDigitalInc/CompressAI/blob/master/compressai/layers/gdn.py

import torch
from torch import nn
import torch.nn.functional as F

from .base import NonNegativeParametrizer

__all__ = [
    "GenDivNorm",
    "InvGenDivNorm"
]

class GenDivNorm(nn.Module):
    r"""Generalized Divisive Normalization layer.
    Introduced in `"Density Modeling of Images Using a Generalized Normalization
    Transformation" <https://arxiv.org/abs/1511.06281>`_,
    by Balle Johannes, Valero Laparra, and Eero P. Simoncelli, (2016).
    .. math::
       y[i] = \frac{x[i]}{\sqrt{\beta[i] + \sum_j(\gamma[j, i] * x[j]^2)}}
    """

    def __init__(self, inChannels: int, groups: int = 1, biasBound: float = 1e-6, weightInit: float = 0.1):
        """Generalized Divisive Normalization layer.

        Args:
            inChannels (int): Channels of input tensor.
            inverse (bool, optional): GDN or I-GDN. Defaults to False.
            beta_min (float, optional): Lower bound of beta. Defaults to 1e-6.
            gamma_init (float, optional): Initial value of gamma. Defaults to 0.1.
        """
        super().__init__()

        self._groups = groups

        biasBound = float(biasBound)
        weightInit = float(weightInit)

        self.beta_reparam = NonNegativeParametrizer(minimum=biasBound)
        beta = torch.ones(inChannels)
        beta = self.beta_reparam.init(beta)
        self.beta = nn.Parameter(beta)

        self.gamma_reparam = NonNegativeParametrizer()
        # m * [cOut // m, cIn // m] -> [cOut, cIn // m]
        gamma = [weightInit * torch.eye(inChannels // self._groups) for _ in range(self._groups)]
        gamma = torch.cat(gamma, 0)
        gamma = self.gamma_reparam.init(gamma)
        self.gamma = nn.Parameter(gamma)

    def forward(self, x):
        # C = x.shape[-3]
        beta = self.beta_reparam(self.beta)
        gamma = self.gamma_reparam(self.gamma)
        # [C, C // groups, 1, 1]
        gamma = gamma[..., None, None]
        std = F.conv2d(x ** 2, gamma, beta, groups=self._groups)

        return self._normalize(x, std)

    def _normalize(self, x: torch.Tensor, std: torch.Tensor) -> torch.Tensor:
        return x * torch.rsqrt(std)


class InvGenDivNorm(GenDivNorm):
    r"""I-GDN layer.
    Introduced in `"Density Modeling of Images Using a Generalized Normalization
    Transformation" <https://arxiv.org/abs/1511.06281>`_,
    by Balle Johannes, Valero Laparra, and Eero P. Simoncelli, (2016).
    .. math::
       y[i] = \frac{x[i]}{\sqrt{\beta[i] + \sum_j(\gamma[j, i] * x[j]^2)}}
    """
    def _normalize(self, x: torch.Tensor, std: torch.Tensor) -> torch.Tensor:
        return x * torch.sqrt(std)


# class EffGenDivNorm(GenDivNorm):
#     r"""Simplified GDN layer.
#     Introduced in `"Computationally Efficient Neural Image Compression"
#     <http://arxiv.org/abs/1912.08771>`_, by Johnston Nick, Elad Eban, Ariel
#     Gordon, and Johannes Ballé, (2019).
#     .. math::
#         y[i] = \frac{x[i]}{\beta[i] + \sum_j(\gamma[j, i] * |x[j]|}
#     """

#     def forward(self, x):
#         C = x.shape[-3]

#         beta = self.beta_reparam(self.beta)
#         gamma = self.gamma_reparam(self.gamma)
#         gamma = gamma.reshape(C, C, 1, 1)
#         norm = F.conv2d(torch.abs(x), gamma, beta)

#         if not self.inverse:
#             norm = 1.0 / norm

#         out = x * norm

#         return out



from typing import Optional
import torch
import torch.nn.functional as F
from torch.utils.data.dataloader import DataLoader
from vlutils.metrics.meter import Meters
from rich.progress import Progress
from mcquic.config import Config

from mcquic.utils.vision import DeTransform
from mcquic.validate.handlers import MsSSIM, PSNR, BPP, IdealBPP, Visualization, ImageCollector
from mcquic.modules.compressor import BaseCompressor


class Validator:
    def __init__(self, config: Config, rank: int):
        self._rank = rank
        self._deTrans = DeTransform().to(rank)
        self._meter = Meters(handlers=[
            MsSSIM().to(rank),
            PSNR().to(rank),
            BPP().to(rank),
            Visualization().to(rank),
            IdealBPP(config.Model.Params["m"], config.Model.Params["k"]).to(rank),
            ImageCollector().to(rank)
        ])

    def tensorToImage(self, x: torch.Tensor) -> torch.Tensor:
        return self._deTrans(x)

    def visualizeIntermediate(self, code: torch.Tensor) -> torch.Tensor:
        code = self._deTrans((code.float() / code.max() - 0.5) * 2)

        # n, m, h, w = code.shape
        # code = code.reshape(n * m, 1, h, w)[:32]

        # only visualize first group
        code = F.interpolate(code[:, :1], scale_factor=4, mode="nearest")
        return code

    @torch.inference_mode()
    def validate(self, epoch: Optional[int], model: BaseCompressor, valLoader: DataLoader, progress: Progress):
        self._meter.reset()
        total = len(valLoader)
        now = 0
        if epoch is None:
            # test mode
            task = progress.add_task(f"[ Test ]", total=total, progress=f"{now:4d}/{total:4d}", suffix="")
        else:
            task = progress.add_task(f"[ Val@{epoch:4d}]", total=total, progress=f"{now:4d}/{total:4d}", suffix="")
        with model.readyForCoding() as cdfs:
            for now, (images, stem) in enumerate(valLoader):
                images = images.to(self._rank, non_blocking=True)
                codes, binaries, headers = model.compress(images, cdfs)
                restored = model.decompress(binaries, cdfs, headers)
                self._meter(images=self.tensorToImage(images), binaries=binaries, restored=self.tensorToImage(restored), codes=codes, stem=stem)
                progress.update(task, advance=1, progress=f"{(now + 1):4d}/{total:4d}")
        progress.remove_task(task)
        return self._meter.results(), self._meter.summary()

    @torch.inference_mode()
    def speed(self, epoch: int, model: BaseCompressor, progress: Progress):
        now = 0
        if epoch is None:
            # test mode
            task = progress.add_task(f"[ Speed test ]", total=100, progress=f"{now:4d}/{100:4d}", suffix="")
        else:
            task = progress.add_task(f"[ Spd@{epoch:4d}]", total=100, progress=f"{now:4d}/{100:4d}", suffix="")

        with model.readyForCoding() as cdfs:
            tensor = torch.rand(10, 3, 768, 512).to(self._rank)

            startEvent = torch.cuda.Event(enable_timing=True)
            endEvent = torch.cuda.Event(enable_timing=True)

            # warm up
            for _ in range(10):
                codes, binaries, headers = model.compress(tensor, cdfs)
                restored = model.decompress(binaries, cdfs, headers)

            startEvent.record()
            for _ in range(50):
                codes, binaries, headers = model.compress(tensor, cdfs)
                progress.update(task, advance=1, progress=f"{(now + 1):4d}/{100:4d}")
                now += 1
            endEvent.record()
            torch.cuda.synchronize()
            encoderMs = startEvent.elapsed_time(endEvent)

            startEvent.record()
            for _ in range(50):
                restored = model.decompress(binaries, cdfs, headers)
                progress.update(task, advance=1, progress=f"{(now + 1):4d}/{100:4d}")
                now += 1
            endEvent.record()
            torch.cuda.synchronize()
            decoderMs = startEvent.elapsed_time(endEvent)

        result = ((50 * 10 * 768 * 512 / 1000) / encoderMs, (50 * 10 * 768 * 512 / 1000) / decoderMs)
        return result, f"Coding throughput: encoder: {result[0]:.2f} Mpps, decoder: {result[1]:.2f} Mpps"



import torch
from torch import nn
import torch.nn.functional as F
from torchvision import transforms as T

from mcquic.utils.vision import RandomHorizontalFlip, RandomVerticalFlip, RandomGamma, RandomAutocontrast

def getTrainingTransform():
    return T.Compose([
        RandomHorizontalFlip(),
        RandomVerticalFlip(),
        RandomAutocontrast(0.25),
        T.ConvertImageDtype(torch.float32),
        RandomGamma(),
        T.Normalize([0.5, 0.5, 0.5], [0.5, 0.5, 0.5]),
    ])

def getTrainingPreprocess():
    return T.Compose([
        T.RandomCrop(512, pad_if_needed=True),
        T.RandomApply([T.RandomChoice([T.ColorJitter(0.4, 0, 0, 0), T.ColorJitter(0, 0.4, 0, 0), T.ColorJitter(0, 0, 0.4, 0), T.ColorJitter(0, 0, 0, 0.2), T.ColorJitter(0.4, 0.4, 0, 0), T.ColorJitter(0.4, 0, 0.4, 0), T.ColorJitter(0.4, 0, 0, 0.2), T.ColorJitter(0, 0.4, 0.4, 0), T.ColorJitter(0, 0.4, 0, 0.2), T.ColorJitter(0, 0, 0.4, 0.2), T.ColorJitter(0.4, 0.4, 0.4, 0), T.ColorJitter(0.4, 0.4, 0, 0.2), T.ColorJitter(0.4, 0, 0.4, 0.2), T.ColorJitter(0, 0.4, 0.4, 0.2), T.ColorJitter(0.4, 0.4, 0.4, 0.2)])], 0.5),
        T.RandomHorizontalFlip(),
        T.RandomVerticalFlip(),
        T.RandomAutocontrast(),
        T.ConvertImageDtype(torch.float32),
        RandomGamma(),
        T.Normalize(0.5, 0.5, True)
    ])

def getEvalTransform():
    return T.Compose([
        T.ConvertImageDtype(torch.float32),
        AlignedCrop(128),
        T.Normalize([0.5, 0.5, 0.5], [0.5, 0.5, 0.5]),
    ])


class AlignedCrop(nn.Module):
    def __init__(self, base: int = 128):
        super().__init__()
        self._base = base

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        h, w = x.shape[-2], x.shape[-1]
        wCrop = w - w // self._base * self._base
        hCrop = h - h // self._base * self._base
        cropLeft = wCrop // 2
        cropRight = wCrop - cropLeft
        cropTop = hCrop // 2
        cropBottom = hCrop - cropTop

        if cropBottom == 0:
            cropBottom = -h
        if cropRight == 0:
            cropRight = -w

        x = x[..., cropTop:(-cropBottom), cropLeft:(-cropRight)]

        return x


class AlignedPadding(nn.Module):
    def __init__(self, base: int = 128):
        super().__init__()
        self._base = base

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        h, w = x.shape[-2], x.shape[-1]
        wPadding = (w // self._base + 1) * self._base - w
        hPadding = (h // self._base + 1) * self._base - h

        wPadding = wPadding % self._base
        hPadding = hPadding % self._base

        padLeft = wPadding // 2
        padRight = wPadding - padLeft
        padTop = hPadding // 2
        padBottom = hPadding - padTop

        return F.pad(x, (padLeft, padRight, padTop, padBottom), "reflect")



"""Package of datasets.

Exports:
    Basic: A Basic dataset that reads all images from a directory.
    BasicLMDB: A Basic dataset that reads from a LMDB.
"""
from typing import Union
import logging
from torch.utils.data import DataLoader, DistributedSampler
from vlutils.logger import LoggerBase
from vlutils.saver import StrPath

from .transforms import getTrainingTransform, getTrainingPreprocess, getEvalTransform
from .dataset import Basic, BasicLMDB


__all__ = [
    "Basic",
    "BasicLMDB",
]


class DummyLoader(DataLoader):
    def __init__(self, **_):
        return

def getTrainLoader(rank: int, worldSize: int, datasetPath: StrPath, batchSize: int, logger: Union[logging.Logger, LoggerBase] = logging.root):
    trainDataset = BasicLMDB(datasetPath, maxTxns=(2 * batchSize) * worldSize, transform=getTrainingPreprocess())
    logger.debug("Create training set: %s", trainDataset)
    trainSampler = DistributedSampler(trainDataset, worldSize, rank)
    trainLoader = DataLoader(trainDataset, batch_size=min(batchSize, len(trainDataset)), sampler=trainSampler, num_workers=2 * batchSize, pin_memory=True, prefetch_factor=4, persistent_workers=True)
    return trainLoader, trainSampler

def getValLoader(datasetPath: StrPath, disable: bool = False, logger: Union[logging.Logger, LoggerBase] = logging.root):
    if disable:
        return DummyLoader()
    valDataset = Basic(datasetPath, transform=getEvalTransform())
    logger.debug("Create validation set: %s", valDataset)
    return DataLoader(valDataset, batch_size=1, shuffle=False, num_workers=4, pin_memory=True, drop_last=False)



from re import I
from typing import Optional, Any, Union, Sequence

from torch import device, Tensor
import torch
from torch.nn import Module
from torch.nn.parallel import DistributedDataParallel

from mcquic.loss import Distortion
from .compressor import BaseCompressor

_device_t = Union[int, device]
_devices_t = Sequence[_device_t]


class _composed(Module):
    def __init__(self, compressor: BaseCompressor, criterion: Distortion) -> None:
        super().__init__()
        self._compressor = compressor
        self._criterion = criterion

    def forward(self, x: Tensor):
        xHat, yHat, codes, logits = self._compressor(x)
        rate, distortion = self._criterion(x, xHat, codes, logits)
        return xHat, (rate, distortion), codes, logits

    def readyForCoding(self):
        return self._compressor.readyForCoding()

    @property
    def Freq(self):
        return self._compressor._quantizer._entropyCoder.Freq

class Composed(DistributedDataParallel):
    def __init__(self, compressor: BaseCompressor, criterion: Distortion, device_ids: Optional[_devices_t] = None, output_device: Optional[_device_t] = None, dim: int = 0, broadcast_buffers: bool = True, process_group: Optional[Any] = None, bucket_cap_mb: float = 25, find_unused_parameters: bool = False, **kwargs):
        module = _composed(compressor, criterion)
        self.module: _composed
        super().__init__(module, device_ids, output_device, dim, broadcast_buffers, process_group, bucket_cap_mb, find_unused_parameters, **kwargs)

    @property
    def Compressor(self):
        return self.module._compressor

    def refresh(self, rank: int) -> torch.Tensor:
        if rank == 0:
            proportion = self.Compressor.reAssignCodebook()
        else:
            proportion = torch.zeros(())
        self.Compressor.syncCodebook()
        return proportion

    def formatDistortion(self, loss: torch.Tensor):
        return self.module._criterion.formatDistortion(loss)



# Copyright 2020 InterDigital Communications, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# https://github.com/InterDigitalInc/CompressAI/blob/master/compressai/layers/layers.py

from typing import Any

import torch
from torch import nn, Tensor


__all__ = [
    "MaskedConv2d",
    "conv3x3",
    "conv5x5",
    "deconv5x5",
    "pixelShuffle5x5",
    "pixelShuffle3x3",
    "conv1x1"
]


class MaskedConv2d(nn.Conv2d):
    """Masked 2D convolution implementation, mask future "unseen" pixels.
    Useful for building auto-regressive network components.
    Introduced in `"Conditional Image Generation with PixelCNN Decoders"
    <https://arxiv.org/abs/1606.05328>`_.
    Inherits the same arguments as a `nn.Conv2d`. Use `maskType='A'` for the
    first layer (which also masks the "current pixel"), `maskType='B'` for the
    following layers.
    """

    def __init__(self, *args: Any, maskType: str = "A", **kwargs: Any):
        """Masked Conv 2D.

        Args:
            args: Positional arguments for `nn.Conv2d`.
            maskType (str, optional): Mask type, if "A", current pixel will be masked, otherwise "B". Use "A" for first layer and "B" for successive layers. Defaults to "A".
            kwargs: Keyword arguments for `nn.Conv2d`.

        Usage:
        ```python
            # First layer
            conv = MaskedConv2d(3, 6, 3, maskType='A')
            # Subsequent layers
            conv = MaskedConv2d(6, 6, 3, maskType='B')
        ```

        Raises:
            ValueError: Mask type not in ["A", "B"].
        """
        super().__init__(*args, **kwargs)

        if maskType not in ("A", "B"):
            raise ValueError(f'Invalid `maskType` value "{maskType}"')

        self.register_buffer("mask", torch.ones_like(self.weight.data))
        _, _, h, w = self.mask.shape
        self.mask[:, :, h // 2, w // 2 + (maskType == "B") :] = 0
        self.mask[:, :, h // 2 + 1 :] = 0

    def forward(self, x: Tensor) -> Tensor:
        # TODO(begaintj): weight assigment is not supported by torchscript
        self.weight.data *= self.mask
        return super().forward(x)

def conv3x3(inChannels: int, outChannels: int, stride: int = 1, bias: bool = True, groups: int = 1) -> nn.Conv2d:
    """A wrapper of 3x3 convolution with pre-calculated padding.

    Usage:
    ```python
        # A 3x3 conv with "same" feature map:
        conv = conv3x3(128, 128)
        # A 3x3 conv with halved feature map:
        conv = conv3x3(128, 128, 2)
    ```

    Args:
        inChannels (int): Channels of input.
        outChannels (int): Channels of output.
        stride (int, optional): Stride. Defaults to 1.
        bias (bool, optional): Bias. Defaults to True.
        groups (int, optional): Group convolution. Defaults to 1.

    Returns:
        nn.Module: A Conv layer.
    """
    return nn.Conv2d(inChannels, outChannels, bias=bias, groups=groups, kernel_size=3, stride=stride, padding=1, padding_mode="reflect")

def conv5x5(inChannels: int, outChannels: int, stride: int = 1, bias: bool = True, groups: int = 1) -> nn.Conv2d:
    """A wrapper of 5x5 convolution with pre-calculated padding.

    Usage:
    ```python
        # A 5x5 conv with "same" feature map:
        conv = conv5x5(128, 128)
        # A 5x5 conv with halved feature map:
        conv = conv5x5(128, 128, 2)
    ```

    Args:
        inChannels (int): Channels of input.
        outChannels (int): Channels of output.
        stride (int, optional): Stride. Defaults to 1.
        bias (bool, optional): Bias. Defaults to True.
        groups (int, optional): Group convolution. Defaults to 1.

    Returns:
        nn.Module: A Conv layer.
    """
    return nn.Conv2d(inChannels, outChannels, bias=bias, groups=groups, kernel_size=5, stride=stride, padding=5 // 2, padding_mode="reflect")

def deconv5x5(inChannels: int, outChannels: int, stride: int = 1, bias: bool = True, groups: int = 1) -> nn.ConvTranspose2d:
    """A wrapper of 5x5 TRANSPOSED convolution with pre-calculated padding.

    #### [CAUTION]: Use it only for proto-tests. Because PyTorch does not use CuDNN when stride > 1 (output_padding will be > 0).
    #### USE `subPixelConv5x5` instead if your want to do 2x up-sampling.

    Usage:
    ```python
        # A 5x5 deconv with "same" feature map:
        conv = deconv5x5(128, 128, 1)
        # A 5x5 deconv with doubled feature map:
        conv = deconv5x5(128, 128)
    ```

    Args:
        inChannels (int): Channels of input.
        outChannels (int): Channels of output.
        stride (int, optional): Stride. Defaults to 1.
        bias (bool, optional): Bias. Defaults to True.
        groups (int, optional): Group convolution. Defaults to 1.

    Returns:
        nn.Module: A Conv layer.
    """
    return nn.ConvTranspose2d(inChannels, outChannels, bias=bias, groups=groups, kernel_size=5, stride=stride, padding=5 // 2, output_padding=stride - 1, padding_mode="zeros")

def pixelShuffle5x5(inChannels: int, outChannels: int, r: float = 1) -> nn.Conv2d:
    """A wrapper of 5x5 convolution and a 2x up-sampling by `PixelShuffle`.

    Usage:
    ```python
        # A 2x up-sampling with a 5x5 conv:
        conv = pixelShuffleConv5x5(128, 128, 2)
        # A 2x down-sampling with a 5x5 conv:
        conv = pixelShuffleConv5x5(128, 128, 0.5)
    ```

    Args:
        inChannels (int): Channels of input.
        outChannels (int): Channels of output.
        stride (int, optional): Stride. Defaults to 1.
        bias (bool, optional): Bias. Defaults to True.
        groups (int, optional): Group convolution. Defaults to 1.

    Returns:
        nn.Module: A Conv layer.
    """
    if r < 1:
        r = int(1 / r)
        return nn.Sequential(
            nn.Conv2d(inChannels, outChannels // (r ** 2), kernel_size=5, stride=1, padding=5 // 2, padding_mode="reflect"),
            nn.PixelUnshuffle(r)
        )
    else:
        r = int(r ** 2)
        return nn.Sequential(
            nn.Conv2d(inChannels, outChannels * r, kernel_size=5, stride=1, padding=5 // 2, padding_mode="reflect"),
            nn.PixelShuffle(r)
        )


def pixelShuffle1x1(inChannels: int, outChannels: int, r: float = 1, groups: int = 1) -> nn.Conv2d:
    """A wrapper of 3x3 convolution and a 2x down-sampling by `PixelShuffle`.

    Usage:
    ```python
        # A 2x down-sampling with a 5x5 conv:
        conv = pixelShuffleConv3x3(128, 128, 0.5)
        # A 2x up-sampling with a 5x5 conv:
        conv = pixelShuffleConv3x3(128, 128, 2)
    ```

    Args:
        inChannels (int): Channels of input.
        outChannels (int): Channels of output.
        stride (int, optional): Stride. Defaults to 1.
        bias (bool, optional): Bias. Defaults to True.
        groups (int, optional): Group convolution. Defaults to 1.

    Returns:
        nn.Module: A Conv layer.
    """
    if r < 1:
        r = int(1 / r)
        return nn.Sequential(
            nn.Conv2d(inChannels, outChannels // (r ** 2), kernel_size=1, groups=groups),
            nn.PixelUnshuffle(r)
        )
    else:
        r = int(r)
        return nn.Sequential(
            nn.Conv2d(inChannels, outChannels * (r ** 2), kernel_size=1, groups=groups),
            nn.PixelShuffle(r)
        )


def pixelShuffle3x3(inChannels: int, outChannels: int, r: float = 1, groups: int = 1) -> nn.Conv2d:
    """A wrapper of 3x3 convolution and a 2x down-sampling by `PixelShuffle`.

    Usage:
    ```python
        # A 2x down-sampling with a 5x5 conv:
        conv = pixelShuffleConv3x3(128, 128, 0.5)
        # A 2x up-sampling with a 5x5 conv:
        conv = pixelShuffleConv3x3(128, 128, 2)
    ```

    Args:
        inChannels (int): Channels of input.
        outChannels (int): Channels of output.
        stride (int, optional): Stride. Defaults to 1.
        bias (bool, optional): Bias. Defaults to True.
        groups (int, optional): Group convolution. Defaults to 1.

    Returns:
        nn.Module: A Conv layer.
    """
    if r < 1:
        r = int(1 / r)
        return nn.Sequential(
            nn.Conv2d(inChannels, outChannels // (r ** 2), kernel_size=3, padding=1, groups=groups, padding_mode="reflect"),
            nn.PixelUnshuffle(r)
        )
    else:
        r = int(r)
        return nn.Sequential(
            nn.Conv2d(inChannels, outChannels * (r ** 2), kernel_size=3, padding=1, groups=groups, padding_mode="reflect"),
            nn.PixelShuffle(r)
        )

def conv1x1(inChannels: int, outChannels: int, stride: int = 1, bias: bool = True, groups: int = 1) -> nn.Conv2d:
    """A wrapper of 1x1 convolution.

    Usage:
    ```python
        # A 1x1 conv with "same" feature map:
        conv = conv1x1(128, 128)
    ```

    Args:
        inChannels (int): Channels of input.
        outChannels (int): Channels of output.
        stride (int, optional): Stride. Defaults to 1.
        bias (bool, optional): Bias. Defaults to True.
        groups (int, optional): Group convolution. Defaults to 1.

    Returns:
        nn.Module: A Conv layer.
    """
    return nn.Conv2d(inChannels, outChannels, bias=bias, groups=groups, kernel_size=1, stride=stride, padding_mode="reflect")



from typing import List, Tuple
import torch
from torch import nn

import mcquic
from mcquic.datasets.transforms import AlignedPadding
from mcquic.nn import pixelShuffle3x3
from mcquic.nn import ResidualBlock, ResidualBlockShuffle, ResidualBlockWithStride
from mcquic.nn.blocks import AttentionBlock
from mcquic.nn.convs import conv3x3
from mcquic.utils.specification import FileHeader, ImageSize

from .quantizer import BaseQuantizer, UMGMQuantizer


class BaseCompressor(nn.Module):
    def __init__(self, encoder: nn.Module, quantizer: BaseQuantizer, decoder: nn.Module):
        super().__init__()
        self._encoder = encoder
        self._decoder = decoder
        self._quantizer = quantizer
        self._qp = "-1"
        self._padding = AlignedPadding()

    @property
    def QuantizationParameter(self) -> str:
        return self._qp

    @QuantizationParameter.setter
    def QuantizationParameter(self, qp: str):
        self._qp = qp

    def forward(self, x: torch.Tensor):
        y = self._encoder(x)
        # [n, c, h, w], [n, m, h, w], [n, m, h, w, k]
        yHat, codes, logits = self._quantizer(y)
        xHat = self._decoder(yHat)
        return xHat, yHat, codes, logits

    def reAssignCodebook(self) -> torch.Tensor:
        return self._quantizer.reAssignCodebook()

    def syncCodebook(self):
        return self._quantizer.syncCodebook()

    def readyForCoding(self):
        return self._quantizer.readyForCoding()

    @property
    def Freq(self):
        return self._quantizer.Freq

    @property
    def CodeUsage(self):
        return torch.cat(list((freq > 0).flatten() for freq in self._quantizer.Freq)).float().mean()

    def compress(self, x: torch.Tensor, cdfs: List[List[List[int]]]) -> Tuple[List[torch.Tensor], List[List[bytes]], List[FileHeader]]:
        n, c, h, w = x.shape

        x = self._padding(x)

        y = self._encoder(x)
        # codes: lv * [n, m, h, w]
        # binaries: List of binary, len = n, len(binaries[0]) = level
        codes, binaries, codeSizes = self._quantizer.compress(y, cdfs)
        header = [FileHeader(mcquic.__version__, self._qp, codeSize, ImageSize(height=h, width=w, channel=c)) for codeSize in codeSizes]
        return codes, binaries, header

    def decompress(self, binaries: List[List[bytes]], cdfs: List[List[List[int]]], headers: List[FileHeader]) -> torch.Tensor:
        yHat = self._quantizer.decompress(binaries, [header.CodeSize for header in headers], cdfs)
        restored = self._decoder(yHat)

        imageSize = headers[0].ImageSize

        H, W = restored.shape[-2], restored.shape[-1]

        h, w = imageSize.height, imageSize.width

        hCrop = H - h
        wCrop = W - w
        cropLeft = wCrop // 2
        cropRight = wCrop - cropLeft
        cropTop = hCrop // 2
        cropBottom = hCrop - cropTop

        if cropBottom == 0:
            cropBottom = -h
        if cropRight == 0:
            cropRight = -w

        return restored[..., cropTop:(-cropBottom), cropLeft:(-cropRight)]


class Compressor(BaseCompressor):
    def __init__(self, channel: int, m: int, k: List[int]):
        encoder = nn.Sequential(
            # convs.conv3x3(3, channel),
            conv3x3(3, channel, 2),
            ResidualBlock(channel, channel, groups=1),
            ResidualBlockWithStride(channel, channel, groups=1),
            AttentionBlock(channel, groups=1),
            ResidualBlock(channel, channel, groups=1),
            ResidualBlockWithStride(channel, channel, groups=1),
            ResidualBlock(channel, channel, groups=1)
        )
        decoder = nn.Sequential(
            ResidualBlock(channel, channel, groups=1),
            ResidualBlockShuffle(channel, channel, groups=1),
            AttentionBlock(channel, groups=1),
            ResidualBlock(channel, channel, groups=1),
            ResidualBlockShuffle(channel, channel, groups=1),
            ResidualBlock(channel, channel, groups=1),
            pixelShuffle3x3(channel, 3, 2)
        )
        quantizer = UMGMQuantizer(channel, m, k, {
            "latentStageEncoder": lambda: nn.Sequential(
                ResidualBlockWithStride(channel, channel, groups=1),
                # GroupSwishConv2D(channel, 3, groups=1),
                ResidualBlock(channel, channel, groups=1),
                AttentionBlock(channel, groups=1),
            ),
            "quantizationHead": lambda: nn.Sequential(
                ResidualBlock(channel, channel, groups=1),
                AttentionBlock(channel, groups=1),
                conv3x3(channel, channel)
                # convs.conv1x1(channel, channel, groups=1)
                # GroupSwishConv2D(channel, channel, groups=1)
            ),
            "latentHead": lambda: nn.Sequential(
                ResidualBlock(channel, channel, groups=1),
                AttentionBlock(channel, groups=1),
                conv3x3(channel, channel)
                # convs.conv1x1(channel, channel, groups=1)
            ),
            "restoreHead": lambda: nn.Sequential(
                AttentionBlock(channel, groups=1),
                ResidualBlock(channel, channel, groups=1),
                ResidualBlockShuffle(channel, channel, groups=1)
            ),
            "dequantizationHead": lambda: nn.Sequential(
                AttentionBlock(channel, groups=1),
                conv3x3(channel, channel),
                ResidualBlock(channel, channel, groups=1),
            ),
            "sideHead": lambda: nn.Sequential(
                AttentionBlock(channel, groups=1),
                conv3x3(channel, channel),
                ResidualBlock(channel, channel, groups=1),
            ),
        })
        super().__init__(encoder, quantizer, decoder)



from typing import Any, Dict, Union, Optional
import os
import logging
import random

import torch
import torch.distributed as dist
import numpy as np
from vlutils.custom import RichProgress
from vlutils.logger import LoggerBase
from vlutils.saver import Saver, DummySaver, StrPath
from vlutils.runtime import functionFullName
from vlutils.base import FrequecyHook
from rich.progress import TimeElapsedColumn, BarColumn, TimeRemainingColumn

from mcquic.utils import nop


def initializeBaseConfigs(port: str, rank: int, worldSize: int, logger: Union[logging.Logger, LoggerBase] = logging.root):
    os.environ["MASTER_ADDR"] = "127.0.0.1"
    os.environ["MASTER_PORT"] = port
    logger.debug("DDP master addr: `%s`", "127.0.0.1")
    logger.debug("DDP master port: `%s`", port)
    torch.autograd.set_detect_anomaly(False)
    torch.backends.cudnn.benchmark = True

    logger.debug("Autograd detect anomaly = `%s`", False)
    logger.debug("         CuDNN bechmark = `%s`", True)
    torch.manual_seed(3407)
    random.seed(3407)
    np.random.seed(3407)
    logger.debug("            Random seed = `%d`", 3407)
    torch.cuda.set_device(rank)
    dist.init_process_group("nccl", world_size=worldSize, rank=rank)
    logger.debug("Process group = `%s`, world size = `%d`", "NCCL", worldSize)


def getRichProgress(disable: bool = False) -> RichProgress:
    return RichProgress("[i blue]{task.description}[/][b magenta]{task.fields[progress]}", TimeElapsedColumn(), BarColumn(None), TimeRemainingColumn(), "{task.fields[suffix]}", refresh_per_second=6, transient=True, disable=disable, expand=True)


def getSaver(saveDir: StrPath, saveName: StrPath = "saved.ckpt", loggerName: str = "root", loggingLevel: Union[str, int] = "INFO", config: Any = None, autoManage: bool = True, maxItems: int = 25, reserve: bool = False, dumpFile: Optional[str] = None, activateTensorboard: bool = True, disable: bool = False):
    if disable:
        return DummySaver(saveDir, saveName, loggerName, loggingLevel, config, autoManage, maxItems, reserve, dumpFile, activateTensorboard)
    else:
        return Saver(saveDir, saveName, loggerName, loggingLevel, config, autoManage, maxItems, reserve, dumpFile, activateTensorboard)

getSaver.__doc__ = Saver.__doc__


def checkHook(function, name, logger: Union[logging.Logger, LoggerBase]=logging.root):
    if function is None:
        logger.debug("No <%s>.", name)
        return nop
    fullName = functionFullName(function)
    logger.debug("<%s> is `%s`.", name, fullName)
    return function


class EpochFrequencyHook(FrequecyHook):
    def __call__(self, step: int, epoch: int, *args: Any, **kwArgs: Any) -> Dict[int, Any]:
        with torch.inference_mode():
            return super().__call__(epoch, step, epoch, *args, **kwArgs)




from collections import OrderedDict

from torch import nn


def build(registry, moduleSpec: OrderedDict):
    layers = list()
    for key, (args, kwArgs) in moduleSpec.items():
        layers.append(
            registry.get(key)(*args, **kwArgs)
        )
    return nn.Sequential(*layers)



from typing import Callable, Optional, Tuple, Callable, List, Union, cast
import os
import json
import sys
from pathlib import Path

import lmdb
import torch
from torch import Tensor
from torchvision.io import read_image
from torchvision.io.image import ImageReadMode, decode_image
from torchvision.datasets import VisionDataset
from torchvision.datasets.folder import IMG_EXTENSIONS, default_loader
from vlutils.runtime import relativePath
from vlutils.types import StrPath


__all__ = [
    "Basic",
    "BasicLMDB"
]


def _hasFileAllowedExtension(filename: str, extensions: Tuple[str, ...]) -> bool:
    """Checks if a file is an allowed extension.

    Args:
        filename (string): path to a file
        extensions (tuple of strings): extensions to consider (lowercase)

    Returns:
        bool: True if the filename ends with one of given extensions
    """
    return filename.lower().endswith(extensions)

def _makeDataset(directory: StrPath, extensions: Optional[Tuple[str, ...]] = None, is_valid_file: Optional[Callable[[str], bool]] = None,) -> List[str]:
    instances = []
    directory = os.path.expanduser(directory)
    both_none = extensions is None and is_valid_file is None
    both_something = extensions is not None and is_valid_file is not None
    if both_none or both_something:
        raise ValueError("Both extensions and is_valid_file cannot be None or not None at the same time")
    def _validFileWrapper(x):
        return _hasFileAllowedExtension(x, cast(Tuple[str, ...], extensions))
    if extensions is not None:
        is_valid_file = _validFileWrapper
    is_valid_file = cast(Callable[[str], bool], is_valid_file)

    for root, _, fnames in sorted(os.walk(directory, followlinks=True)):
        for fname in sorted(fnames):
            path = os.path.join(root, fname)
            if is_valid_file(path):
                instances.append(path)
    return instances


class Basic(VisionDataset):
    """A Basic dataset that reads all images from a directory.
    """
    def __init__(self, root: StrPath, transform: Optional[Callable] = None, is_valid_file: Optional[Callable[[str], bool]] = None) -> None:
        """A Basic dataset that reads all images from a directory.

        Usage:
        ```python
            dataset = Basic("data/clic/train", T.Compose([
                                                   T.ToTensor(),
                                                   T.Normalize(mean, std)]))
        ```

        Args:
            root (str): The folder to read from.
            duplicate (int, optional): A number that repeat images in this dataset for N times. Defaults to 1.
            transform (Optional[Callable], optional): A transform applies to images. Defaults to None.
            is_valid_file (Optional[Callable[[str], bool]], optional): A function that check image is valid. Defaults to None, will use builtin implementation.

        Raises:
            RuntimeError: Find no images in this folder.
        """
        super().__init__(root, transform=transform)
        samples = _makeDataset(self.root, IMG_EXTENSIONS if is_valid_file is None else None, is_valid_file)
        if len(samples) == 0:
            msg = "Found 0 files in subfolders of: {}\n".format(self.root)
            msg += "Supported extensions are: {}".format(",".join(IMG_EXTENSIONS))
            raise RuntimeError(msg)
        self.root = root
        self.loader = default_loader
        self.extensions = IMG_EXTENSIONS
        self.samples = samples

    def __getitem__(self, index: int) -> Tuple[Tensor, str]:
        """
        Args:
            index (int): Index

        Returns:
            Tensor: sample at index.
        """
        path = self.samples[index]
        # No need to force RGB. Transforms will handle it.
        sample = read_image(path, ImageReadMode.UNCHANGED)
        if self.transform is not None:
            sample = self.transform(sample)
        return sample, Path(path).stem

    def __len__(self) -> int:
        return len(self.samples)

    def __str__(self) -> str:
        return f"<Basic> at `{relativePath(self.root)}` with transform: \r\n`{self.transform}`"


class BasicLMDB(VisionDataset):
    """A Basic dataset that reads from a LMDB.
    """
    def __init__(self, root: StrPath, maxTxns: int = 1, repeat: int = 1, transform: Optional[Callable] = None) -> None:
        """A Basic dataset that reads from a LMDB.

        Usage:
        ```python
            dataset = Basic("data/trainSet", numWorkers + 2,
                               transform=T.Compose([T.ToTensor(),
                                                    T.Normalize(mean, std)]))
        ```

        Args:
            root (str): LMDB folder path. In addition, a `meatadata.json` should be placed in the same folder --- see `src/misc/datasetCreate.py`
            maxTxns (int, optional): Max trasactions of LMDB. Defaults to 1.
            repeat (int, optional): Repeat images of the dataset for N times. Defaults to 1.
            transform (Optional[Callable], optional): Transform applies to images. Defaults to None.
        """
        super().__init__(root, transform=transform)
        self._root = root
        self._maxTxns = maxTxns
        # env and txn is lazy-loaded in ddp. They can't be pickled
        self._env: Union[lmdb.Environment, None] = None
        self._txn: Union[lmdb.Transaction, None] = None
        # Length is needed for DistributedSampler, but we can't use env to get it, env can't be pickled.
        # So we decide to read from metadata placed in the same folder --- see src/misc/datasetCreate.py
        with open(os.path.join(root, "metadata.json"), "r") as fp:
            metadata = json.load(fp)
        self._length = metadata["length"]
        self._repeat = repeat

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._txn is not None:
            self._txn.__exit__(exc_type, exc_val, exc_tb)
        if self._env is not None:
            self._env.close()

    def _initEnv(self):
        self._env = lmdb.open(self.root, map_size=int(1024 ** 4), subdir=True, readonly=True, readahead=False, meminit=False, max_spare_txns=self._maxTxns, lock=False)
        self._txn = self._env.begin(write=False, buffers=True)

    def __getitem__(self, index: int) -> Tensor:
        """
        Args:
            index (int): Index

        Returns:
            Tensor: sample at index.
        """
        index = index % self._length
        if self._env is None or self._txn is None:
            self._initEnv()
        sample = torch.ByteTensor(torch.ByteStorage.from_buffer(bytearray(self._txn.get(index.to_bytes(32, sys.byteorder)))))
        # UNCHANGED --- Slightly speedup
        # No need to force RGB. Transforms will handle it.
        sample = decode_image(sample, ImageReadMode.UNCHANGED)
        if sample.shape[0] == 1:
            sample = sample.repeat((3, 1, 1))
        elif sample.shape[0] == 4:
            sample = sample[:3]
        if self.transform is not None:
            sample = self.transform(sample)
        return sample

    def __len__(self) -> int:
        return self._length * self._repeat

    def __str__(self) -> str:
        return f"<LMDB> at `{relativePath(self.root)}` with transform: \r\n`{self.transform}`"



import json
import os
from random import shuffle
import shutil
import sys
import pathlib

import lmdb
from PIL import Image
import PIL
import click
from mcquic.train.utils import getRichProgress

from mcquic.utils import hashOfFile

_EXT = [".png", ".jpg", ".jpeg"]


def write(txn, i: bytes, path: str):
    with open(path, "rb") as fp:
        txn.put(i, fp.read())


def findAllWithSize(dirPath, ext):
    files = list()
    filesAndDirs = os.listdir(dirPath)
    for f in filesAndDirs:
        if os.path.isdir(os.path.join(dirPath, f)):
            files.extend(findAllWithSize(os.path.join(dirPath, f), ext))
        elif os.path.splitext(f)[1].lower() in ext:
            f = os.path.join(dirPath, f)
            files.append((f, os.path.getsize(f)))
    return files


def getFilesFromDir(root, progress, strict: bool = False):
    files = findAllWithSize(root, _EXT)
    newFile = list()

    total = len(files)

    task = progress.add_task(f"[ Check ]", total=total, progress="0.00%", suffix=f"Scanning {total} images")

    for i, f in enumerate(files):
        try:
            a = Image.open(f[0])
        except PIL.UnidentifiedImageError:
            continue
        w, h = a.size
        if strict:
            # force images size > 512.
            if h < 512 or w < 512:
                continue
        newFile.append(f)

        progress.update(task, advance=1, progress=f"{(i + 1) / total * 100 :.2f}%")

    print(f"{len(newFile)} images meets requirement.")\

    progress.remove_task(task)

    return [x[0] for x in newFile]


def main(imageFolder: pathlib.Path, targetDir: pathlib.Path):
    shutil.rmtree(targetDir, ignore_errors=True)
    os.makedirs(targetDir, exist_ok=True)

    progress = getRichProgress()

    with progress:

        allFiles = getFilesFromDir(imageFolder, progress, True)

        shuffle(allFiles)

        env = lmdb.Environment(str(targetDir), subdir=True, map_size=int(1024 ** 4))

        total = len(allFiles)

        task = progress.add_task(f"[ Write ]", total=total, progress="0.00%", suffix=f"Writing {total} images")

        with env.begin(write=True) as txn:
            i = -1
            for i, f in enumerate(allFiles):
                write(txn, i.to_bytes(32, sys.byteorder), f)
                progress.update(task, advance=1, progress=f"{(i + 1) / total * 100 :.2f}%")
        env.close()

        progress.remove_task(task)


        dbFile = os.path.join(targetDir, "data.mdb")

        print("Calculate database hash...")

        hashResult = hashOfFile(dbFile, progress)

        # Create metadata needed for dataset
        with open(os.path.join(targetDir, "metadata.json"), "w") as fp:
            json.dump({
                "length": i + 1,
                "hash": hashResult
            }, fp)


CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
@click.command(context_settings=CONTEXT_SETTINGS)
@click.option("-D", "--debug", is_flag=True, help="Set logging level to DEBUG to print verbose messages.")
@click.option("-q", "--quiet", is_flag=True, help="Silence all messages, this option has higher priority to `-D/--debug`.")
@click.argument("images", type=click.Path(exists=True, file_okay=False, resolve_path=True, path_type=pathlib.Path), required=True, nargs=1)
@click.argument("output", type=click.Path(exists=False, file_okay=False, resolve_path=True, path_type=pathlib.Path), required=True, nargs=1)
def entryPoint(debug, quiet, path, images, output):
    """Create training set from `images` dir to `output` dir.

Args:

    images (str): All training images folder, allow sub-folders.

    output (str): Output dir to create training set.
    """
    main(images, output)




import logging
import os
import shutil
import tempfile
import atexit


class ConstsMetaClass(type):
    @property
    def TempDir(cls):
        if getattr(cls, '_tempDir', None) is None:
            tempDir = os.path.dirname(tempfile.mktemp())
            tempDir = os.path.join(tempDir, "mcquic")
            cls._tempDir = tempDir
            os.makedirs(cls._tempDir, exist_ok=True)
            def removeTmp():
                shutil.rmtree(tempDir, ignore_errors=True)
            atexit.register(removeTmp)
        return cls._tempDir

class Consts(metaclass=ConstsMetaClass):
    Name = "mcquic"
    # lazy load
    # TempDir = "/tmp/mcquic/"
    Eps = 1e-6
    CDot = "·"
    TimeOut = 30



# Copyright 2020 by Gongfan Fang, Zhejiang University.
# All rights reserved.
from math import sqrt
import warnings

import torch
from torch import nn
import torch.nn.functional as F

__all__ = [
    "MsSSIM",
    "Ssim",
    "ms_ssim",
    "ssim",
    "psnr",
    "PSNR"
]

_WEIGHTS = [0.0448, 0.2856, 0.3001, 0.2363, 0.1333]


def _fspecial_gauss_1d(size, sigma, device=None):
    r"""Create 1-D gauss kernel
    Args:
        size (int): the size of gauss kernel
        sigma (float): sigma of normal distribution

    Returns:
        torch.Tensor: 1D kernel (1 x 1 x size)
    """
    coords = torch.arange(size, device=device).float()
    coords -= size // 2

    g = torch.exp(-(coords ** 2) / (2 * sigma ** 2))
    g /= g.sum()

    return g.unsqueeze(0).unsqueeze(0)


def _gaussian_filter(x, win):
    r""" Blur x with 1-D kernel
    Args:
        x (torch.Tensor): a batch of tensors to be blurred
        window (torch.Tensor): 1-D gauss kernel

    Returns:
        torch.Tensor: blurred tensors
    """
    assert all(ws == 1 for ws in win.shape[1:-1]), win.shape
    if len(x.shape) == 4:
        conv = F.conv2d
    elif len(x.shape) == 5:
        conv = F.conv3d
    else:
        raise NotImplementedError(x.shape)

    C = x.shape[1]
    out = x
    for i, s in enumerate(x.shape[2:]):
        if s >= win.shape[-1]:
            out = conv(out, weight=win.transpose(2 + i, -1), stride=1, padding=0, groups=C)
        else:
            warnings.warn(
                f"Skipping Gaussian Smoothing at dimension 2+{i} for input: {x.shape} and win size: {win.shape[-1]}"
            )

    return out


def _ssim(X, Y, data_range, win, K=(0.01, 0.03)):

    r""" Calculate ssim index for X and Y

    Args:
        X (torch.Tensor): images
        Y (torch.Tensor): images
        win (torch.Tensor): 1-D gauss kernel
        data_range (float or int, optional): value range of input images. (usually 1.0 or 255)

    Returns:
        torch.Tensor: ssim results.
    """
    K1, K2 = K
    # batch, channel, [depth,] height, width = X.shape
    compensation = 1.0

    C1 = (K1 * data_range) ** 2
    C2 = (K2 * data_range) ** 2

    mu1 = _gaussian_filter(X, win)
    mu2 = _gaussian_filter(Y, win)

    mu1_sq = mu1.pow(2)
    mu2_sq = mu2.pow(2)
    mu1_mu2 = mu1 * mu2

    sigma1_sq = compensation * (_gaussian_filter(X * X, win) - mu1_sq)
    sigma2_sq = compensation * (_gaussian_filter(Y * Y, win) - mu2_sq)
    sigma12 = compensation * (_gaussian_filter(X * Y, win) - mu1_mu2)

    cs_map = (2 * sigma12 + C2) / (sigma1_sq + sigma2_sq + C2)  # set alpha=beta=gamma=1
    ssim_map = ((2 * mu1_mu2 + C1) / (mu1_sq + mu2_sq + C1)) * cs_map

    ssim_per_channel = torch.flatten(ssim_map, 2).mean(-1)
    cs = torch.flatten(cs_map, 2).mean(-1)
    return ssim_per_channel, cs


def ssim(X, Y, win, data_range=255, sizeAverage=True, K=(0.01, 0.03), nonnegative_ssim=False):
    r""" interface of ssim
    Args:
        X (torch.Tensor): a batch of images, (N,C,H,W)
        Y (torch.Tensor): a batch of images, (N,C,H,W)
        data_range (float or int, optional): value range of input images. (usually 1.0 or 255)
        sizeAverage (bool, optional): if sizeAverage=True, ssim of all images will be averaged as a scalar
        win_size: (int, optional): the size of gauss kernel
        win_sigma: (float, optional): sigma of normal distribution
        win (torch.Tensor, optional): 1-D gauss kernel. if None, a new kernel will be created according to win_size and win_sigma
        K (list or tuple, optional): scalar constants (K1, K2). Try a larger K2 constant (e.g. 0.4) if you get a negative or NaN results.
        nonnegative_ssim (bool, optional): force the ssim response to be nonnegative with relu

    Returns:
        torch.Tensor: ssim results
    """
    assert len(X.shape) in (4, 5), f"Input images should be 4-d or 5-d tensors, but got {X.shape}"

    assert X.type() == Y.type(), "Input images should have the same dtype."

    win_size = win.shape[-1]

    assert win_size % 2 == 1, "Window size should be odd."

    ssim_per_channel, _ = _ssim(X, Y, data_range=data_range, win=win, K=K)
    if nonnegative_ssim:
        ssim_per_channel = torch.relu(ssim_per_channel)

    if sizeAverage:
        return ssim_per_channel.mean()
    return ssim_per_channel.mean(1)


def ms_ssim(X, Y, win, weights, poolMethod, data_range=255, sizeAverage=True, K=(0.01, 0.03)):

    r""" interface of ms-ssim
    Args:
        X (torch.Tensor): a batch of images, (N,C,[T,]H,W)
        Y (torch.Tensor): a batch of images, (N,C,[T,]H,W)
        data_range (float or int, optional): value range of input images. (usually 1.0 or 255)
        sizeAverage (bool, optional): if sizeAverage=True, ssim of all images will be averaged as a scalar
        win_size: (int, optional): the size of gauss kernel
        win_sigma: (float, optional): sigma of normal distribution
        win (torch.Tensor, optional): 1-D gauss kernel. if None, a new kernel will be created according to win_size and win_sigma
        weights (Tensor, optional): weights for different levels
        K (list or tuple, optional): scalar constants (K1, K2). Try a larger K2 constant (e.g. 0.4) if you get a negative or NaN results.
    Returns:
        torch.Tensor: ms-ssim results
    """
    assert X.shape == Y.shape, "Input images should have the same dimensions."

    # for d in range(len(X.shape) - 1, 1, -1):
    #     X = X.squeeze(dim=d)
    #     Y = Y.squeeze(dim=d)

    win_size = win.shape[-1]
    assert win_size % 2 == 1, "Window size should be odd."

    smaller_side = min(X.shape[-2:])
    assert smaller_side > (win_size - 1) * (
        2 ** 4
    ), "Image size should be larger than %d due to the 4 downsamplings in ms-ssim" % ((win_size - 1) * (2 ** 4))

    levels = weights.shape[0]
    mcs = []
    ssim_per_channel = None
    for i in range(levels):
        ssim_per_channel, cs = _ssim(X, Y, win=win, data_range=data_range, K=K)

        if i < levels - 1:
            mcs.append(torch.relu(cs))
            padding = [s % 2 for s in X.shape[2:]]
            X = poolMethod(X, kernel_size=2, padding=padding)
            Y = poolMethod(Y, kernel_size=2, padding=padding)

    if ssim_per_channel is None:
        raise ValueError()

    ssim_per_channel = torch.relu(ssim_per_channel)  # (batch, channel)
    mcs_and_ssim = torch.stack(mcs + [ssim_per_channel], dim=1)  # (batch, level, channel)
    ms_ssim_val = torch.prod(mcs_and_ssim ** weights.view(1, -1, 1), dim=1)

    if sizeAverage:
        return ms_ssim_val.mean()
    return ms_ssim_val.mean(1)


class Ssim(nn.Module):
    def __init__(self, data_range=255, sizeAverage=True, win_size=11, win_sigma=1.5, channel=3, spatial_dims=2, K=(0.01, 0.03), nonnegative_ssim=False):
        r""" class for ssim
        Args:
            data_range (float or int, optional): value range of input images. (usually 1.0 or 255)
            sizeAverage (bool, optional): if sizeAverage=True, ssim of all images will be averaged as a scalar
            win_size: (int, optional): the size of gauss kernel
            win_sigma: (float, optional): sigma of normal distribution
            channel (int, optional): input channels (default: 3)
            K (list or tuple, optional): scalar constants (K1, K2). Try a larger K2 constant (e.g. 0.4) if you get a negative or NaN results.
            nonnegative_ssim (bool, optional): force the ssim response to be nonnegative with relu.
        """

        super().__init__()
        self.register_buffer("win", _fspecial_gauss_1d(win_size, win_sigma).repeat([channel, 1] + [1] * spatial_dims))
        self.sizeAverage = sizeAverage
        self.data_range = data_range
        self.K = K
        self.nonnegative_ssim = nonnegative_ssim

    def forward(self, X, Y):
        return 1.0 - ssim(X, Y, self.win, data_range=self.data_range, sizeAverage=self.sizeAverage, K=self.K, nonnegative_ssim=self.nonnegative_ssim)


class MsSSIM(nn.Module):
    def __init__(self, shape=4, data_range=255, sizeAverage=True, win_size=11, win_sigma=1.5, channel=3, spatial_dims=2, weights=None, K=(0.01, 0.03)):
        r""" class for ms-ssim
        Args:
            shape (int): 4 for NCHW, 5 for NCTHW
            data_range (float or int, optional): value range of input images. (usually 1.0 or 255)
            sizeAverage (bool, optional): if sizeAverage=True, ssim of all images will be averaged as a scalar
            win_size: (int, optional): the size of gauss kernel
            win_sigma: (float, optional): sigma of normal distribution
            channel (int, optional): input channels (default: 3)
            weights (list, optional): weights for different levels
            K (list or tuple, optional): scalar constants (K1, K2). Try a larger K2 constant (e.g. 0.4) if you get a negative or NaN results.
        """

        super().__init__()

        if shape == 4:
            self._avg_pool = F.avg_pool2d
        elif shape == 5:
            self._avg_pool = F.avg_pool3d
        else:
            raise ValueError(f"Input shape should be 4-d or 5-d tensors, but got {shape}")

        self.register_buffer("win", _fspecial_gauss_1d(win_size, win_sigma).repeat([channel, 1] + [1] * spatial_dims), persistent=False)
        self.sizeAverage = sizeAverage
        self.data_range = data_range
        if weights is None:
            weights = torch.tensor(_WEIGHTS)
        else:
            weights = torch.tensor(weights)
        self.register_buffer("weights", weights, persistent=False)
        self.K = K

    def forward(self, X, Y):
        return 1.0 - ms_ssim(X, Y, self.win, self.weights, self._avg_pool, data_range=self.data_range, sizeAverage=self.sizeAverage, K=self.K)


def psnr(x: torch.Tensor, y: torch.Tensor, sizeAverage: bool = False, upperBound: float = 255.0):
    mse = ((x.float() - y.float()) ** 2).mean(dim=(1, 2, 3))
    res = 10 * (upperBound ** 2 / mse).log10()
    return res.mean() if sizeAverage else res


class PSNR(nn.Module):
    def __init__(self, sizeAverage: bool = False, upperBound: float = 255.0):
        super().__init__()
        self.register_buffer("_upperBound", torch.tensor(upperBound ** 2))
        self._average = sizeAverage

    def forward(self, x: torch.Tensor, y: torch.Tensor):
        mse = ((x.float() - y.float()) ** 2).mean(dim=(1, 2, 3))
        res = 10 * (self._upperBound / mse).log10()
        return res.mean() if self._average else res



import logging
import pathlib
import random

import torch
import torch.multiprocessing as mp
import click
import yaml
from vlutils.runtime import queryGPU

from mcquic.config import Config
from .ddp import ddpSpawnTraining, registerForTrain

def checkArgs(debug, quiet, resume: pathlib.Path, configPath: pathlib.Path):
    if resume is None and configPath is None:
        raise ValueError("Both `resume` and `config` are not given.")
    if quiet:
        return logging.CRITICAL
    if debug:
        return logging.DEBUG
    return logging.INFO

def main(debug: bool, quiet: bool, resume: pathlib.Path, configPath: pathlib.Path) -> int:
    assert False, "You need to run `mcquic train` with Python optimized-mode. Try re-run me with `python -O -m mcquic.train ...`"
    loggingLevel = checkArgs(debug, quiet, resume, configPath)

    logging.getLogger().setLevel(loggingLevel)

    if configPath is not None:
        config = Config.deserialize(yaml.full_load(configPath.read_text()))
        logging.debug("Use fresh config.")
    elif resume is not None:
        ckpt = torch.load(resume, "cpu")
        config = Config.deserialize(ckpt["config"])
        logging.debug("Use saved config.")
    else:
        raise ValueError("Both `--resume` and `config` are None.")

    gpus = queryGPU(needGPUs=config.Train.GPU.GPUs, wantsMore=config.Train.GPU.WantsMore, needVRamEachGPU=(config.Train.GPU.VRam + 256) if config.Train.GPU.VRam > 0 else -1, writeOSEnv=True)
    worldSize = len(gpus)

    masterPort = str(random.randint(10001, 65535))

    registerForTrain()

    # `daemon` is True --- Way to handle SIGINT globally.
    # Give up handling SIGINT by yourself... PyTorch hacks it.
    mp.spawn(ddpSpawnTraining, (worldSize, masterPort, config, config.Train.SaveDir, resume, loggingLevel), worldSize, daemon=True)
    return 0


CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
@click.command(context_settings=CONTEXT_SETTINGS)
@click.option("-D", "--debug", is_flag=True, help="Set logging level to DEBUG to print verbose messages.")
@click.option("-q", "--quiet", is_flag=True, help="Silence all messages, this option has higher priority to `-D/--debug`.")
@click.option("-r", "--resume", type=click.Path(exists=True, dir_okay=False, resolve_path=True, path_type=pathlib.Path), required=False, nargs=1, help="`.ckpt` file path to resume training.")
@click.argument('config', type=click.Path(exists=True, dir_okay=False, resolve_path=True, path_type=pathlib.Path), required=False, nargs=1)
def entryPoint(debug, quiet: bool, resume: pathlib.Path, config: pathlib.Path):
    main(debug, quiet, resume, config)



from contextlib import contextmanager
from typing import List, Tuple, Generator

import torch
from torch import nn
import torch.distributed as dist

# Port from CompressAI
from mcquic.rans import pmfToQuantizedCDF
from mcquic.rans import RansEncoder, RansDecoder

from mcquic.utils.specification import CodeSize


class EntropyCoder(nn.Module):
    def __init__(self, m: int, k: List[int], ema: float = 0.9):
        super().__init__()
        self.encooder = RansEncoder()
        self.decoder = RansDecoder()
        # initial value is uniform
        self._freqEMA = nn.ParameterList(nn.Parameter(torch.ones(m, ki), requires_grad=False) for ki in k)
        self._k = k
        self._decay = 1 - ema

    @torch.no_grad()
    def forward(self, oneHotCodes: List[torch.Tensor]):
        # Update freq by EMA
        # [n, m, h, w, k]
        for lv, code in enumerate(oneHotCodes):
            # [m, k]
            totalCount = code.sum((0, 2, 3))
            # sum over all gpus
            dist.all_reduce(totalCount)

            # up trigger don't perform EMA
            rmsMask = totalCount > self._freqEMA[lv]
            # otherwise
            ema = self._decay * totalCount + (1 - self._decay) * self._freqEMA[lv]

            # update
            self._freqEMA[lv].copy_(totalCount * rmsMask + ema * ~rmsMask)

    @contextmanager
    def readyForCoding(self) -> Generator[List[List[List[int]]], None, None]:
        cdfs = list()
        for freq in self.Freq:
            cdfAtLv = list()
            for freqAtM in freq:
                total = freqAtM.sum()
                if total < 1:
                    prob = torch.ones_like(freqAtM, dtype=torch.float) / len(freqAtM)
                else:
                    prob = freqAtM.float() / total
                cdf = pmfToQuantizedCDF(prob.tolist(), 16)
                cdfAtLv.append(cdf)
            cdfs.append(cdfAtLv)
        try:
            yield cdfs
        finally:
            del cdfs

    @property
    def Freq(self):
        """Yields list of `[m, k]` tensors.

        Yields:
            torch.Tensor: `[m, k]` tensor of level `l`.
        """
        for freqEMA in self._freqEMA:
            # normalize with precision 16bits.
            yield (freqEMA / freqEMA.sum(-1, keepdim=True) * 65536).round().long()

    def _checkShape(self, codes: List[torch.Tensor]):
        info = "Please give codes with correct shape, for example, [[1, 2, 24, 24], [1, 2, 12, 12], ...], which is a `level` length list. each code has shape [n, m, h, w]. "
        if len(codes) < 1:
            raise RuntimeError("Length of codes is 0.")
        n = codes[0].shape[0]
        m = codes[0].shape[1]
        for code in codes:
            newN, newM = code.shape[0], code.shape[1]
            if n < 1:
                raise RuntimeError(info + "Now `n` = 0.")
            if m != newM:
                raise RuntimeError(info + "Now `m` is inconsisitent.")
            if n != newN:
                raise RuntimeError(info + "Now `n` is inconsisitent.")
        return n, m

    @torch.inference_mode()
    def compress(self, codes: List[torch.Tensor], cdfs: List[List[List[int]]]) -> Tuple[List[List[bytes]], List[CodeSize]]:
        """Compress codes to binary.

        Args:
            codes (List[torch.Tensor]): List of tensor, len = level, code.shape = [n, m, h, w]
            cdfs (List[List[List[int]]]): cdfs for entropy coder, len = level, len(cdfs[0]) = m

        Returns:
            List[List[bytes]]: List of binary, len = n, len(binary[0]) = level
            List[CodeSize]]: List of code size, len = n
        """
        n, m = self._checkShape(codes)
        compressed = list(list() for _ in range(n))
        heights = list()
        widths = list()
        # [n, m, h, w]
        for code, ki, cdf in zip(codes, self._k, cdfs):
            _, _, h, w = code.shape
            heights.append(h)
            widths.append(w)
            for i, codePerImage in enumerate(code):
                indices = torch.arange(m)[:, None, None]
                # [m, h, w]
                idx = indices.expand_as(codePerImage).flatten().int().tolist()
                cdfSizes = [ki + 2] * m
                # [m, h, w]
                offsets = torch.zeros_like(codePerImage).flatten().int().tolist()
                binary: bytes = self.encooder.encode_with_indexes(codePerImage.flatten().int().tolist(), idx, cdf, cdfSizes, offsets)
                compressed[i].append(binary)
        return compressed, [CodeSize(m, heights, widths, self._k) for _ in range(n)]

    @torch.inference_mode()
    def decompress(self, binaries: List[List[bytes]], codeSizes: List[CodeSize], cdfs: List[List[List[int]]]) -> List[torch.Tensor]:
        """Restore codes from binary

        Args:
            binaries (List[List[bytes]]): len = n, len(binary[0]) = level
            codeSizes (List[CodeSize]): len = n
            cdfs (List[List[List[int]]]): len = level, len(cdfs[0]) = m

        Returns:
            List[List[torch.Tensor]]: len = level, each code.shape = [n, m, h, w]
        """
        lv = len(binaries[0])
        m = codeSizes[0].m
        codes = list(list() for _ in range(lv))
        indices = torch.arange(m)[:, None, None]
        for binary, codeSize in zip(binaries, codeSizes):
            for lv, (binaryAtLv, cdf, ki, h, w) in enumerate(zip(binary, cdfs, self._k, codeSize.heights, codeSize.widths)):
                idx = indices.expand(codeSize.m, h, w).flatten().int().tolist()
                cdfSizes = [ki + 2] * codeSize.m
                offsets = torch.zeros(codeSize.m, h, w, dtype=torch.int).flatten().int().tolist()
                restored: List[int] = self.decoder.decode_with_indexes(binaryAtLv, idx, cdf, cdfSizes, offsets)
                # [m, h, w]
                code = torch.tensor(restored).reshape(codeSize.m, h, w)
                codes[lv].append(code)
        return [torch.stack(c, 0).to(self._freqEMA[0].device) for c in codes]



import click

from .stCompressService import main


CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
@click.command(context_settings=CONTEXT_SETTINGS)
@click.option("-D", "--debug", is_flag=True, help="Set logging level to DEBUG to print verbose messages.")
@click.option("-q", "--quiet", is_flag=True, help="Silence all messages, this option has higher priority to `-D/--debug`.")
@click.option("-qp", type=click.IntRange(1, 13), default=2, show_default=True, help="Quantization parameter. Higher means better image quality and larger size.")
@click.option("-d", "--disable-gpu", is_flag=True, help="Disable GPU usage.")
def entryPoint(debug, quiet, qp, disable_gpu):
    """Start a Streamlit service for compressing and restoring images
    """
    raise NotImplementedError



import os
from pathlib import Path
from setuptools import setup

from pybind11.setup_helpers import Pybind11Extension, build_ext, ParallelCompile

ParallelCompile("NPY_NUM_BUILD_JOBS").install()

CWD = Path("./")
PKG_NAME = "mcquic"

# https://github.com/InterDigitalInc/CompressAI/blob/master/setup.py
# Copyright (c) 2021-2022, InterDigital Communications, Inc
def get_extensions():
    ext_dirs = CWD / "third_party/CompressAI/cpp_exts"
    ext_modules = []

    # Add rANS module
    rans_lib_dir = CWD / "third_party/CompressAI/ryg_rans"

    if os.name == "nt":
        compiler_args = ["/std:c++17", "/O2", "/GL", "/MP8"]
    else:
        compiler_args = ["-std=c++17", "-O3"]

    ext_modules.append(
        Pybind11Extension(
            name=f"{PKG_NAME}.rans",
            sources=[str(s) for s in ext_dirs.glob("*.cpp")],
            language="c++",
            include_dirs=[rans_lib_dir, ext_dirs],
            extra_compile_args=compiler_args,
        )
    )

    return ext_modules


setupArgs = {
    "ext_modules": get_extensions(),
    "cmdclass": {
        "build_ext": build_ext
    }
}

if "PYPI_BUILDING" in os.environ:
    install_requires = [
        "torch>=1.10,<2",
        "torchvision>=0.11,<1",
        "tensorboard>=2.3,<3",
        "rich<11",
        "lmdb<2",
        "pyyaml<7",
        "marshmallow<4",
        "click<9",
        "vlutils",
        "msgpack-python<2",
        # "streamlit<2",
        "packaging"
    ]

    setupArgs.update({
        "install_requires": install_requires
    })

if "ADD_ENTRY" in os.environ:
    console_scripts = [
        "mcquic = mcquic.cli:entryPoint",
        "mcquic-train = mcquic.train.cli:entryPoint",
        "mcquic-dataset = mcquic.datasets.cli:entryPoint",
        "mcquic-validate = mcquic.validate.cli:entryPoint"
    ]
    setupArgs.update({
        "entry_points": {
            'console_scripts': console_scripts
        }
    })


setup(**setupArgs)



import pathlib

import click
import torch
import torch.hub
from vlutils.utils import DefaultGroup

import mcquic



def version(ctx, _, value):
    if not value or ctx.resilient_parsing:
        return
    click.echo(r"""
                                       :::::.:::.
#####*       +#####                  ::         .=   .:...+=   -...== -...-+    ........=-
%%%%%%+     -%%%%%%                 -     -=.    -#  :    %*  :    ##.:   =%  .:        #+
%%%%%%%:   .#%%%%%%     :=+***+-   -    -#*-+    .%: :   -%:  -   :%=-    #+ :.   .=+++*%:
%%%%#%%#.  #%%#%%%%  .+%%%%%%%%*  -    -%=  =    :%-:    ##  ::   *%.-   -%.-    =%+:
%%%%=%%%* +%%%=%%%% .#%%%%+::--   -   .%*...+    +%.:   -%=  =   .%=:    #*.:   -%-
%%%%-=%%%=%%%-=%%%% +%%%%*       :.   -*    .   .%*..   *%  :.   +% -   -%--    #+
%%%%= *%%%%%* +%%%% +%%%%+       -    :%=       *# :    #* .:   :%=::   *# -    #-
%%%%=  #%%%#  *%%%% :%%%%%-  .-  .:    -=      *%: :    -+:.   .#* =   .%= -    -=::::
%%%%=  :%%%.  *%%%%  -#%%%%%%%%*  -           .#:   -         -%* .:   :%. .:        *+
****=         =****    -+*#%##*=.  :::...:-:   .+    :-::.:-+%*-  -   .+%   .:-:....-%:
                                     .-====--====*.    :-===:     -.-*%+:      :======
                                             :::::.              :##+:
""" + mcquic.__version__)
    ctx.exit()


CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
@click.group(cls=DefaultGroup, context_settings=CONTEXT_SETTINGS)
@click.option("-v", "--version", is_flag=True, callback=version, expose_value=False, is_eager=True, help="Print version info.")
def entryPoint():
    pass


@entryPoint.command(default=True)
@click.option("-D", "--debug", is_flag=True, help="Set logging level to DEBUG to print verbose messages.")
@click.option("-q", "--quiet", is_flag=True, help="Silence all messages, this option has higher priority to `-D/--debug`.")
@click.option("-qp", type=click.IntRange(0, 13), default=2, show_default=True, help="Quantization parameter. Higher means better image quality and larger size.")
@click.option("--local", type=click.Path(exists=True, dir_okay=False, resolve_path=True, path_type=pathlib.Path), help="Use a local model path instead of download by `qp`.")
@click.option("--disable-gpu", is_flag=True, help="Use pure CPU to perform compression. This will be slow.")
@click.option("--mse", is_flag=True, help="Use model optimized for PSNR other than MsSSIM.")
@click.option("--crop", is_flag=True, help="Crop the image to align feature patches. Edges of image are cutted though, compressed binary will be smaller.")
@click.argument('input', type=click.Path(exists=True, dir_okay=False, resolve_path=True, path_type=pathlib.Path), nargs=1)
@click.argument('output', type=click.Path(exists=False, resolve_path=True, path_type=pathlib.Path), required=False, nargs=1)
def _(debug, quiet, qp, local, disable_gpu, mse, crop, input, output):
    """Compress/restore a file.

Args:

    input (str): Input file path. If input is an image, compress it. If input is a `.mcq` file, restore it.

    output (optional, str): Output file path or dir. If not provided, this program will only print compressor information of input file.
    """
    from .demo import main
    with torch.inference_mode():
        main(debug, quiet, qp, local, disable_gpu, mse, crop, input, output)


@entryPoint.command()
@click.option("-D", "--debug", is_flag=True, help="Set logging level to DEBUG to print verbose messages.")
@click.option("-q", "--quiet", is_flag=True, help="Silence all messages, this option has higher priority to `-D/--debug`.")
@click.option("-r", "--resume", type=click.Path(exists=True, dir_okay=False, resolve_path=True, path_type=pathlib.Path), required=False, nargs=1, help="`.ckpt` file path to resume training.")
@click.argument('config', type=click.Path(exists=True, dir_okay=False, resolve_path=True, path_type=pathlib.Path), required=False, nargs=1)
def train(debug, quiet, resume, config):
    """Train a model.

Args:

    config (str): Config file (yaml) path. If `-r/--resume` is present but config is still given, then this config will be used to update the resumed training.
    """
    from mcquic.train.cli import main
    main(debug, quiet, resume, config)


@entryPoint.command()
@click.option("-D", "--debug", is_flag=True, help="Set logging level to DEBUG to print verbose messages.")
@click.option("-q", "--quiet", is_flag=True, help="Silence all messages, this option has higher priority to `-D/--debug`.")
@click.option("-e", "--export", type=click.Path(exists=False, resolve_path=True, path_type=pathlib.Path), required=False, help="Path to export the final model that is compatible with main program.")
@click.argument("path", type=click.Path(exists=True, dir_okay=False, resolve_path=True, path_type=pathlib.Path), required=True, nargs=1)
@click.argument("images", type=click.Path(exists=True, file_okay=False, resolve_path=True, path_type=pathlib.Path), required=True, nargs=1)
@click.argument("output", type=click.Path(exists=True, file_okay=False, resolve_path=True, path_type=pathlib.Path), required=False, nargs=1)
def validate(debug, quiet, export, path, images, output):
    """Validate a trained model from `path` by images from `images` dir, and publish a final state_dict to `output` path.

Args:

    path (str): Saved checkpoint path.

    images (str): Validation images folder.

    output (str): Dir to save all restored images.
    """
    from mcquic.validate.cli import main
    with torch.inference_mode():
        main(debug, quiet, export, path, images, output)


@entryPoint.command()
@click.option("-D", "--debug", is_flag=True, help="Set logging level to DEBUG to print verbose messages.")
@click.option("-q", "--quiet", is_flag=True, help="Silence all messages, this option has higher priority to `-D/--debug`.")
@click.argument("images", type=click.Path(exists=True, file_okay=False, resolve_path=True, path_type=pathlib.Path), required=True, nargs=1)
@click.argument("output", type=click.Path(exists=False, file_okay=False, resolve_path=True, path_type=pathlib.Path), required=True, nargs=1)
def dataset(debug, quiet, images, output):
    """Create training set from `images` dir to `output` dir.

Args:

    images (str): All training images folder, allow sub-folders.

    output (str): Output dir to create training set.
    """
    from mcquic.datasets.cli import main
    main(images, output)



from typing import Union, List, Tuple

import torch
from torch import nn

class Decibel(nn.Module):
    def __init__(self, upperBound: float) -> None:
        super().__init__()
        self._upperBound = upperBound ** 2

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return -10 * (x / self._upperBound).log10()


class EMATracker(nn.Module):
    def __init__(self, size: Union[torch.Size, List[int], Tuple[int, ...]], momentum: float = 0.9):
        super().__init__()
        self._shadow: torch.Tensor
        self._decay = 1 - momentum
        self.register_buffer("_shadow", torch.empty(size) * torch.nan)

    @torch.no_grad()
    def forward(self, x: torch.Tensor):
        if torch.all(torch.isnan(self._shadow)):
            self._shadow.copy_(x)
            return self._shadow
        self._shadow -= self._decay * (self._shadow - x)
        return self._shadow
